import { workerFactory } from './worker-interface.js';
import { isSchedulingFailure } from './api-types.js';

/**
 * Runs (in a separate thread) the list scheduling algorithm on the given problem instance and returns the result
 * asynchronously.
 *
 * See [the project page](https://github.com/fschopp/project-planning-js) for more information on the algorithm.
 *
 * @param args argument list that will be passed on to {@link computeSchedule}() unaltered
 * @return promise that will be resolved with the solution or rejected with a {@link SchedulingFailure} containing a
 *     human-readable failure description if the problem instance is invalid (for example, has a cyclic dependency
 *     graph)
 */

function computeScheduleAsync(...args) {
  const worker = workerFactory.createWorker();

  if (isSchedulingFailure(worker)) {
    const failure = worker;
    return Promise.reject(failure);
  }
  // Hence all worker callbacks are in place before we send it the "go" message below.

  const promise = new Promise((resolve, reject) => {
    worker.onmessage = event => {
      const result = event.data;

      if (isSchedulingFailure(result)) {
        reject(result);
      } else {
        resolve(result);
      }
    };

    worker.onerror = event => {
      worker.terminate();
      const failure = 'A runtime error occurred in source file ' + `${event.filename} (line ${event.lineno}:${event.colno}):\n${event.message}`; // In theory, the worker could still cause an error after it has sent its last message. However, the ES6
      // specification says:
      // "Attempting to resolve or reject a resolved promise has no effect."
      // https://www.ecma-international.org/ecma-262/6.0/#sec-promise-objects
      // Hence, there is no race condition. On the other hand, we don't want to swallow the problem. Therefore, we
      // add an assertion:

      reject(failure);
    };
  });
  worker.postMessage(args);
  return promise;
}

export { computeScheduleAsync };
//# sourceMappingURL=scheduling-async.js.map
