import { JobSplitting } from './api-types.js';
import MinHeap from './minheap.js';

/**
 * Algorithm for project planning. More precisely, *list scheduling* with support for machines running at different
 * speeds, optional preemption, optional splitting of jobs across machines, release dates, and delivery times.
 *
 * See README file for details.
 */
const JOB_DEFAULTS = Object.freeze({
  deliveryTime: 0,
  splitting: JobSplitting.PREEMPTION,
  dependencies: [],
  releaseTime: 0,
  preAssignment: -1
});
/**
 * Runs the list scheduling algorithm on the given problem instance and returns the result.
 *
 * See [the project page](https://github.com/fschopp/project-planning-js) for more information on the algorithm.
 *
 * @param instance the problem instance
 * @return solution or a human-readable failure description if the problem instance is invalid (for example, has a
 *     cyclic dependency graph)
 */

function computeSchedule(instance) {
  const nonNegativeInteger = number => number >= 0 && Number.isInteger(number);

  const positiveIntegers = numbers => numbers.filter(number => number <= 0 || !Number.isInteger(number)).length === 0;

  const undefinedOrNonNegativeInteger = number => number === undefined || nonNegativeInteger(number);

  const undefinedOrNonNegativeIntegers = (...numbers) => numbers.filter(number => !undefinedOrNonNegativeInteger(number)).length === 0;

  const undefinedOrIntegerWithinZeroTo = (upperBoundExcl, number) => number === undefined || nonNegativeInteger(number) && number < upperBoundExcl;

  const integersWithinZeroTo = (upperBoundExcl, numbers) => numbers.filter(number => !(number >= 0 && nonNegativeInteger(number) && number < upperBoundExcl)).length === 0;

  const numMachines = instance.machineSpeeds.length;
  const numJobs = instance.jobs.length;

  if (numMachines === 0) {
    return 'At least one machine is required to compute a schedule.';
  } else if (!positiveIntegers(instance.machineSpeeds) || instance.jobs.filter(job => !nonNegativeInteger(job.size) || !undefinedOrNonNegativeIntegers(job.deliveryTime, job.releaseTime) || !undefinedOrIntegerWithinZeroTo(numMachines, job.preAssignment) || job.dependencies !== undefined && !integersWithinZeroTo(numJobs, job.dependencies)).length > 0 || !undefinedOrNonNegativeInteger(instance.minFragmentSize)) {
    return 'All job processing times and machine speeds need to be non-negative integers. ' + 'All job dependency and pre-assignment indices need to be within bounds.';
  }

  const jobs = instance.jobs.map(job => Object.assign({}, JOB_DEFAULTS, job));
  const minFragmentSize = instance.minFragmentSize === undefined ? 0 : instance.minFragmentSize;
  return new ListScheduling(numMachines, instance.machineSpeeds, jobs, minFragmentSize).schedule;
}
/**
 * Internal class whose sole purpose is to store algorithm state, so it does not have to be inconveniently passed around
 * between functions.
 */

class ListScheduling {
  constructor(numMachines, machineSpeeds, jobs, minFragmentSize) {
    this.numMachines_ = numMachines;
    this.machineSpeeds_ = machineSpeeds;
    this.jobs_ = jobs;
    this.minFragmentSize_ = minFragmentSize;
    this.gapsLists_ = this.machineSpeeds_.map(ignoredMachineSpeed => ({
      head: {
        startTime: Number.MIN_SAFE_INTEGER,
        endTime: 0
      },
      tail: {
        head: {
          startTime: 0,
          endTime: Number.MAX_SAFE_INTEGER
        },
        tail: null
      }
    }));
    this.schedule = this.computeSchedule();
  }
  /**
   * Returns the next “event” for the given machine.
   *
   * This method has side-effects. It updates field {@link Machine.previousGapsList} and {@link Machine.gapsList} of the
   * given machine.
   */


  static nextTimeStamp(minWallClockTime, earliestStart, machine) {
    let currentGap = machine.gapsList.head;

    if (machine.currentFragmentStart === null) {
      while (true) {
        const timeStamp = Math.max(earliestStart, currentGap.startTime) + minWallClockTime;

        if (timeStamp <= currentGap.endTime) {
          return timeStamp;
        }

        machine.previousGapsList = machine.gapsList;
        machine.gapsList = machine.gapsList.tail;
        currentGap = machine.gapsList.head;
      }
    } else {
      return currentGap.endTime;
    }
  }
  /**
   * Adjust the linked list of gaps to account for the newly scheduled job fragment.
   *
   * There are 4 cases to consider. The new job fragment either:
   * 1. Fills out the current gap completely. In this case, the current gap needs to be removed
   *    entirely.
   * 2. Aligns with the start of the current gap, but ends before it. In this case, the start of the current gap is
   *    changed to when the new job fragment ends.
   * 3. Does not start with the current gap, but aligns with the end of it. In this case, the end of the current gap is
   *    changed to when the job fragment starts.
   * 4. If the current gap aligns with neither start nor end of the current gap, a new gap is inserted before the
   *    current gap. It starts with the current gap and ends with the start of the job fragment. The start of the
   *    current gap is then changed to the end of the job fragment.
   *
   * @param machineState State of the machine that the job fragment has been scheduled on.
   * @param end end time of the job fragment
   * @param commitGaps if true, then commit the updated gaps to the state that is shared across scheduling of individual
   *     jobs
   */


  adjustGaps(machineState, end, commitGaps) {
    // machineState also points to state that is shared across scheduling of individual jobs. This shared state we must
    // not modify if dryRun is true. We therefore cannot make modifications to the properties of
    // machineState.previousGapsList or machineState.gap.
    const start = machineState.currentFragmentStart;
    const currentGap = machineState.gapsList.head;

    if (commitGaps) {
      if (start === currentGap.startTime && end === currentGap.endTime) {
        machineState.previousGapsList.tail = machineState.gapsList.tail;
      } else if (start === currentGap.startTime) {
        currentGap.startTime = end;
      } else if (end === currentGap.endTime) {
        currentGap.endTime = start;
        machineState.previousGapsList = machineState.gapsList; // Note that machineState.gapsList will be updated below.
      } else {
        machineState.previousGapsList.tail = {
          head: {
            startTime: currentGap.startTime,
            endTime: start
          },
          tail: machineState.gapsList
        };
        currentGap.startTime = end;
      }
    }

    if (end === currentGap.endTime) {
      // Note that machineState.gapsList.tail === null would imply that the current gap is the last gap; that is,
      // currentGap.endTime === Number.MAX_SAFE_INTEGER. But we are also in the case where (end === currentGap.endTime).
      // Since we don't support schedules that long, it is safe to assert:
      machineState.gapsList = machineState.gapsList.tail;
    } else if (!commitGaps) {
      machineState.gapsList = {
        head: {
          startTime: end,
          endTime: currentGap.endTime
        },
        tail: machineState.gapsList.tail
      };
    } // Upon return, we will be "before" the gap represented by machineState.gapsList.head.


    machineState.currentFragmentStart = null;
  }

  createJobFragment(machineState, end, scheduledJob) {
    if (scheduledJob) {
      const start = machineState.currentFragmentStart;

      if (end - start > 0) {
        const jobFragment = {
          machine: machineState.index,
          start,
          end,
          isWaiting: false
        };
        scheduledJob.push(jobFragment);
      }
    }

    this.adjustGaps(machineState, end, scheduledJob !== undefined);
  }
  /**
   * Schedules processing of a job and returns its completion time (excluding delivery time).
   */


  scheduleJob(availableMachineIndices, size, isPreemptible, earliestStart, scheduledJob) {
    const minFragmentSize = isPreemptible ? Math.min(size, this.minFragmentSize_) : size;
    let currentSpeed = 0;
    let lastTimestamp = earliestStart;
    let remainingSize = size;
    const machines = availableMachineIndices.map((index, availableIdx) => {
      const gapsList = this.gapsLists_[index];
      return {
        index,
        availableIdx,
        speed: this.machineSpeeds_[index],
        currentFragmentStart: null,
        previousGapsList: gapsList,
        gapsList: gapsList.tail
      };
    });

    while (remainingSize > 0) {
      let eventTime = Number.MAX_SAFE_INTEGER;
      let machine = machines[0];
      availableMachineIndices.forEach((index, availableIdx) => {
        const minWallClockTime = Math.ceil(minFragmentSize / machines[availableIdx].speed);
        const machineEventTime = ListScheduling.nextTimeStamp(minWallClockTime, earliestStart, machines[availableIdx]);

        if (machineEventTime < eventTime) {
          eventTime = machineEventTime;
          machine = machines[availableIdx];
        }
      });
      let isProjectedEndTime = false;
      const currentProjectedEndTime = Math.ceil(lastTimestamp + remainingSize / currentSpeed);

      if (currentProjectedEndTime < eventTime) {
        eventTime = currentProjectedEndTime;
        isProjectedEndTime = true;
      }

      remainingSize -= (eventTime - lastTimestamp) * currentSpeed;

      if (!isProjectedEndTime) {
        if (machine.currentFragmentStart === null) {
          const nextGap = machine.gapsList.head;
          machine.currentFragmentStart = Math.max(earliestStart, nextGap.startTime);
          remainingSize -= (eventTime - machine.currentFragmentStart) * machine.speed;
          currentSpeed += machine.speed;
        } else {
          this.createJobFragment(machine, eventTime, scheduledJob);
          currentSpeed -= machine.speed;
        }
      }

      lastTimestamp = eventTime;
    }

    availableMachineIndices.forEach((ignoredIndex, availableIdx) => {
      if (machines[availableIdx].currentFragmentStart !== null) {
        this.createJobFragment(machines[availableIdx], lastTimestamp, scheduledJob);
      }
    });
    return lastTimestamp;
  }
  /**
   * Adds the given delivery time on the given machine to the given job schedule, and returns the overall completion
   * time of the job.
   */


  static scheduleDeliveryTime(machineIdx, processingCompletionTime, deliveryTime, scheduledJob) {
    let completionTime = processingCompletionTime;

    if (deliveryTime > 0) {
      completionTime += deliveryTime;
      const deliveryJobFragment = {
        machine: machineIdx,
        start: processingCompletionTime,
        end: completionTime,
        isWaiting: true
      };
      scheduledJob.push(deliveryJobFragment);
    }

    return completionTime;
  }

  allMachines() {
    return {
      length: this.machineSpeeds_.length,

      forEach(callbackfn) {
        for (let i = 0; i < this.length; ++i) {
          callbackfn(i, i);
        }
      },

      map(callbackfn) {
        const array = [];
        array.length = this.length;

        for (let i = 0; i < this.length; ++i) {
          array[i] = callbackfn(i, i);
        }

        return array;
      }

    };
  }

  static singleMachine(singleMachineIdx) {
    return {
      length: 1,

      forEach(callbackfn) {
        callbackfn(singleMachineIdx, 0);
      },

      map(callbackfn) {
        return [callbackfn(singleMachineIdx, 0)];
      }

    };
  }

  computeSchedule() {
    const noDependencyNodes = [];
    const jobGraphNodes = this.jobs_.map((job, index) => ({
      idx: index,
      numDependencies: job.dependencies.length,
      dependents: []
    }));

    for (let i = 0; i < this.jobs_.length; ++i) {
      const job = this.jobs_[i];
      const jobGraphNode = jobGraphNodes[i];

      for (const dependencyIdx of job.dependencies) {
        jobGraphNodes[dependencyIdx].dependents.push(jobGraphNode);
      }

      if (job.dependencies.length === 0) {
        noDependencyNodes.push(jobGraphNode);
      }
    }

    const allMachines = this.allMachines();
    const newSchedule = this.jobs_.map(ignoredJob => []);
    const finishTimes = this.jobs_.map(ignoredJob => -1);
    let numScheduledJobs = 0;
    const noDependencyNodesHeap = new MinHeap(noDependencyNodes, (left, right) => left.idx - right.idx);

    while (!noDependencyNodesHeap.isEmpty()) {
      const jobGraphNode = noDependencyNodesHeap.extractMin();
      const job = this.jobs_[jobGraphNode.idx];
      const isPreemptible = job.splitting !== JobSplitting.NONE;
      const earliestStartTime = job.dependencies.reduce((previousEarliestStartTime, dependencyIdx) => {
        const dependencyFinishTime = finishTimes[dependencyIdx];
        return Math.max(previousEarliestStartTime, dependencyFinishTime);
      }, job.releaseTime);
      let availableMachines = allMachines;
      let deliveryMachineIdx;

      if (job.splitting === JobSplitting.MULTIPLE_MACHINES) {
        deliveryMachineIdx = job.preAssignment >= 0 ? job.preAssignment : 0;
      } else if (job.preAssignment >= 0) {
        availableMachines = ListScheduling.singleMachine(job.preAssignment);
        deliveryMachineIdx = job.preAssignment;
      } else {
        let maxCompletionTime = Number.MAX_SAFE_INTEGER;

        for (let i = 0; i < this.numMachines_; ++i) {
          const currentAvailableMachines = ListScheduling.singleMachine(i);
          const currentCompletionTime = this.scheduleJob(currentAvailableMachines, job.size, isPreemptible, earliestStartTime);

          if (currentCompletionTime < maxCompletionTime) {
            availableMachines = currentAvailableMachines;
            deliveryMachineIdx = i;
            maxCompletionTime = currentCompletionTime;
          }
        }
      }

      const completionTime = this.scheduleJob(availableMachines, job.size, isPreemptible, earliestStartTime, newSchedule[jobGraphNode.idx]);
      finishTimes[jobGraphNode.idx] = ListScheduling.scheduleDeliveryTime(deliveryMachineIdx, completionTime, job.deliveryTime, newSchedule[jobGraphNode.idx]);
      ++numScheduledJobs;

      for (const dependent of jobGraphNode.dependents) {
        --dependent.numDependencies;

        if (dependent.numDependencies === 0) {
          noDependencyNodesHeap.add(dependent);
        }
      }
    }

    return numScheduledJobs !== this.jobs_.length ? 'Detected a cycle in the dependency graph.' : newSchedule;
  }

}

export { computeSchedule };
//# sourceMappingURL=scheduling.js.map
